#!/usr/bin/env python

"""
File access.

Copyright (C) 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.terms import *
from os import listdir, remove, rename  # partition manipulation
from shutil import copy                 # index updating
from os.path import join

try:
    set
except NameError:
    from sets import Set as set

# Constants.

TERM_FILENAMES    = "terms", "term-index"

# Utility functions.

def get_partitions(pathname, prefix):

    """
    Return a set of partition names for partitions residing at the given
    'pathname' having the given 'prefix'.
    """

    prefix_length = len(prefix)

    partitions = set()
    for filename in listdir(pathname):
        if filename.startswith(prefix):
            partition = filename[prefix_length:]
            if partition.isdigit():
                partitions.add(int(partition))
    return partitions

def get_term_partitions(pathname):

    """
    Return a set of term partition identifiers for partitions residing at the
    given 'pathname'.
    """

    return get_partitions(pathname, "terms-")

def get_next_partition(partitions):
    return max(partitions or [-1]) + 1

def get_writer(pathname, name, partition, cls):
    f = open(join(pathname, "%s-%s" % (name, partition)), "wb")
    return cls(f)

def get_term_writer(pathname, partition):

    """
    Return a term writer using files under the given 'pathname' labelled
    according to the given 'partition'.
    """

    return get_writer(pathname, "terms", partition, TermWriter)

def get_term_index_writer(pathname, partition):

    """
    Return a term index writer using files under the given 'pathname' labelled
    according to the given 'partition'.
    """

    return get_writer(pathname, "term-index", partition, TermIndexWriter)

def get_reader(pathname, name, partition, cls):
    f = open(join(pathname, "%s-%s" % (name, partition)), "rb")
    return cls(f)

def get_term_reader(pathname, partition):

    """
    Return a term reader using files under the given 'pathname' labelled
    according to the given 'partition'.
    """

    return get_reader(pathname, "terms", partition, TermIterator)

def get_term_data_reader(pathname, partition):

    """
    Return a term plus data reader using files under the given 'pathname'
    labelled according to the given 'partition'.
    """

    return get_reader(pathname, "terms", partition, TermDataIterator)

def get_term_index_reader(pathname, partition):

    """
    Return a term index reader using files under the given 'pathname' labelled
    according to the given 'partition'.
    """

    return get_reader(pathname, "term-index", partition, TermIndexIterator)

def get_combined_term_reader(pathname, partition):

    """
    Return a combined term reader using files under the given 'pathname'
    labelled according to the given 'partition'.
    """

    return CombinedIterator(get_term_reader(pathname, partition), get_term_index_reader(pathname, partition))

# Renaming.

def rename_files(pathname, names, from_partition, to_partition):
    for name in names:
        try:
            rename(join(pathname, "%s-%s" % (name, from_partition)), join(pathname, "%s-%s" % (name, to_partition)))
        except OSError:
            pass

def rename_term_files(pathname, from_partition, to_partition):
    rename_files(pathname, TERM_FILENAMES, from_partition, to_partition)

# Removal/deletion.

def remove_files(pathname, names, partition):
    for name in names:
        try:
            remove(join(pathname, "%s-%s" % (name, partition)))
        except OSError:
            pass

def remove_term_files(pathname, partition):
    remove_files(pathname, TERM_FILENAMES, partition)

# Copying.

def copy_files(source, names, partition, destination, suffix):
    for name in names:
        filename = "%s-%s" % (name, partition)
        try:
            copy(join(source, filename), join(destination, filename + suffix))
        except OSError:
            pass

def copy_term_files(source, partition, destination, suffix):
    copy_files(source, TERM_FILENAMES, partition, destination, suffix)

# vim: tabstop=4 expandtab shiftwidth=4
