#!/usr/bin/env python

"""
Specific classes for storing term information.

Copyright (C) 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.data import *
from iixr.files import *
from itermerge import itermerge
from os.path import commonprefix # to find common string prefixes
from bisect import bisect_right, insort_right
import operator

class TermWriter(FileWriter):

    "Writing term information to files."

    def begin(self, docnum_size, position_size):

        "Begin writing to the file."

        self.write_numbers((docnum_size, position_size))
        self.end_record()

        self.data_start = self.tell()
        self.docnum_size = docnum_size
        self.position_size = position_size
        self.subtractor = get_subtractor(docnum_size)
        self.last_term = ""

    def write_terms(self, terms):

        """
        Write the 'terms' to the term information file, with each term's details
        stored in a separate record.
        """

        if hasattr(terms, "items"):
            terms = terms.items()
            terms.sort()

        for term, doc_positions in terms:
            if not doc_positions:
                continue

            if hasattr(doc_positions, "items"):
                doc_positions = doc_positions.items()

            docnum, positions = doc_positions[0]

            if not positions:
                continue

            # Start the writing, if appropriate.

            if self.data_start is None:
                self.begin(sizeof(docnum), sizeof(positions[0]))

            # Write each term and document positions.

            self.write_term(term, doc_positions)
            self.end_record()

    # Methods requiring an open record.

    def write_term(self, term, doc_positions):

        """
        Write the given 'term', its document frequency (number of documents in
        which it appears), and 'doc_positions' to the term information file.
        """

        self.write_term_only(term)

        # Write the document frequency and the term positions.

        self.write_positions(doc_positions)

    def write_term_plus_remaining(self, term, data):

        "Write the given 'term' and the document position 'data'."

        self.write_term_only(term)
        self.write_remaining(data)

    def write_term_only(self, term):

        "Write only the given 'term'."

        if term <= self.last_term:
            raise ValueError, "Term %r precedes the previous term %r." % (term, self.last_term)

        # Write the prefix length and term suffix.

        common = len(commonprefix([self.last_term, term]))
        suffix = term[common:]

        self.write_number(common)
        self.write_string(suffix)

        self.last_term = term

    def write_positions(self, doc_positions):

        "Write the given 'doc_positions' to the file."

        # Make sure that the positions are sorted.

        doc_positions.sort()

        # Write the document frequency.

        self.write_number(len(doc_positions))

        last_docnum = None

        for docnum, positions in doc_positions:

            # Store the first document number as it is.

            if last_docnum is None:
                docnum_seq = docnum

            # Reject out-of-order documents.

            elif docnum < last_docnum:
                raise ValueError, "Document number %r is less than previous number %r." % (docnum, last_docnum)

            # Calculate an ongoing delta.

            else:
                docnum_seq = self.subtractor(docnum, last_docnum)

            # Write the document number and positions.

            self.write_sequence_value(docnum_seq, self.docnum_size)
            self.write_monotonic_sequence(positions, self.position_size)

            last_docnum = docnum

        # Write a terminating byte to indicate that no more document pages
        # exist.

        self.write_byte(0)

class TermReader(FileReader):

    "Reading term information from files."

    def begin(self):

        "Begin reading from the file."

        self.begin_record()
        try:
            self.docnum_size, self.position_size = self.read_numbers(2)
        except EOFError:
            self.docnum_size, self.position_size = 0, 0 # NOTE: No positions!

        self.data_start = self.tell()
        self.adder = get_adder(self.docnum_size)
        self.last_term = ""

    def get_sizes(self):
        return self.docnum_size, self.position_size

    # Methods requiring an open record.

    def read_term(self):

        "Read a term and its document positions from the term information file."

        # Read the term.

        self.read_term_only()

        # Read the document frequency and the term positions.

        positions = self.read_positions()

        return self.last_term, positions

    def read_term_plus_remaining(self):

        """
        Read a term and the unprocessed document position data.
        """

        self.read_term_only()
        return self.last_term, self.read_remaining()

    def read_term_only(self):

        "Read a term only."

        # Read the prefix length and term suffix.

        common = self.read_number()
        suffix = self.read_string()

        self.last_term = self.last_term[:common] + suffix
        return self.last_term

    def read_positions(self):

        "Read document positions from the term information file."

        doc_positions = []

        while 1:

            # Read the document frequency.

            npositions = self.read_number()

            last_docnum = None
            i = 0
            while i < npositions:

                # Read the document number.

                docnum = self.read_sequence_value(self.docnum_size)
                if last_docnum is not None:
                    docnum = self.adder(docnum, last_docnum)

                # Read the positions.

                positions = self.read_monotonic_sequence(self.position_size)
                doc_positions.append((docnum, positions))

                last_docnum = docnum
                i += 1

            # Read a terminating byte to discover whether more document pages
            # exist.

            if not self.read_byte():
                break

        return doc_positions

# Indexes covering the information files.

class TermIndexWriter(FileWriter):

    "Writing term index information to files."

    def begin(self):

        "Begin writing to the file."

        self.data_start = self.tell()
        self.last_term = ""
        self.last_offset = 0

    def write_term(self, term, offset):

        "Write the given 'term' and 'offset'."

        if term <= self.last_term:
            raise ValueError, "Term %r precedes the previous term %r." % (term, self.last_term)

        # Write the prefix length and term suffix.

        common = len(commonprefix([self.last_term, term]))
        suffix = term[common:]

        self.write_number(common)
        self.write_string(suffix)

        # Write the offset delta.

        self.write_number(offset - self.last_offset)

        self.last_term = term
        self.last_offset = offset

class TermIndexReader(FileReader):

    "Reading term index information to files."

    def begin(self):

        "Begin reading from the file."

        self.data_start = self.tell()
        self.last_term = ""
        self.last_offset = 0

    def read_term(self):

        "Read a term and an offset from the file."

        # Read the prefix length and term suffix.

        common = self.read_number()
        suffix = self.read_string()

        self.last_term = self.last_term[:common] + suffix

        # Read the offset delta.

        self.last_offset += self.read_number()
        return self.last_term, self.last_offset

# External reading classes.

class TermIterator(TermReader):

    "An iterator over terms and positions read from a file."

    def __iter__(self):
        return self

    def next(self):
        try:
            self.begin_record()
            return self.read_term()
        except EOFError:
            raise StopIteration

class TermDataIterator(TermReader):

    "An iterator over terms and unprocessed document positions data."

    def __iter__(self):
        return self

    def next(self):
        try:
            self.begin_record()
            return self.read_term_plus_remaining()
        except EOFError:
            raise StopIteration

class TermIndexIterator(TermIndexReader):

    "An iterator over terms and offsets read from a file."

    def __iter__(self):
        return self

    def next(self):
        try:
            self.begin_record()
            return self.read_term()
        except EOFError:
            raise StopIteration

class CombinedIterator:

    "An iterator providing index and information file access."

    def __init__(self, reader, index_reader):
        self.reader = reader
        self.index_reader = index_reader
        self.records = list(index_reader)
        self.terms = [t for t, dp in self.records]

        # Cache the last term and positions.

        self.last_term_returned = None
        self.last_positions_returned = None

    def go_to_term(self, term):

        """
        Return the 'term' and positions or nearest following term and positions.
        """

        if self.last_term_returned is not None and self.last_term_returned == term:
            return self.last_term_returned, self.last_positions_returned

        # Get the record providing a term less than or equal to the requested
        # term, getting the first entry if no such records exist.

        after = bisect_right(self.terms, term)
        before = max(0, after - 1)

        t, offset = self.records[before]
        terms_after = self.terms[after:]

        # Seek to the corresponding record in the information file.
        # Only do this if the term is more quickly reached by seeking. 

        if term <= t or (
            self.last_term_returned is None or term < self.last_term_returned or
            self.last_term_returned < t or terms_after and terms_after[0] <= self.last_term_returned
            ):

            self.reader.seek(offset)

            # Skip the term information, overwrite the reader's state.

            self.reader.begin_record()
            self.reader.read_term_only()
            self.reader.last_term = t

        # Where the found term is equal or greater, just read the positions for
        # the index entry.

        if t >= term:

            # Get the positions.

            self.last_term_returned, self.last_positions_returned = t, self.reader.read_positions()
            return self.last_term_returned, self.last_positions_returned

        # Where the found term is less, use the information file to find the
        # term or the one after.

        else:

            # Scan for the term.

            while t < term:
                t, dp = self.next() # remembers the term and positions

            return t, dp

    def __iter__(self):
        return self

    def next(self):
        self.last_term_returned, self.last_positions_returned = record = self.reader.next()
        return record

    def close(self):
        if self.reader is not None:
            self.reader.close()
            self.reader = None
        if self.index_reader is not None:
            self.index_reader.close()
            self.index_reader = None

class MultipleReader(itermerge):

    """
    Accessing many term readers at once.
    NOTE: This does not currently work with CombinedIterator instances because
    NOTE: iteration causes the next method to be called in advance, and this
    NOTE: obstructs interleaving of go_to_term and next invocations.
    """

    def __init__(self, readers, combine=None):

        """
        Initialise a master index reader using underlying 'readers' and a
        'combine' function which knows how to combine position information from
        different sources.
        """

        self.readers = readers
        self.combine = combine or operator.add

        # Initialise this object as an iterator over the readers.

        itermerge.__init__(self, self.readers)
        self.next_value = None

    def get_sizes(self):

        # Readers must have compatible sizes.

        if self.readers:
            return self.readers[0].get_sizes()
        else:
            return 0, 0

    def go_to_term(self, term):

        # Refresh the iterators to provide sought values.

        self.iters = []
        for reader in self.readers:
            try:
                insort_right(self.iters, (reader.go_to_term(term), reader.next))
            except StopIteration:
                pass

        self.next_value = None
        return self.next()

    def next(self):

        # Return any prematurely read value.

        if self.next_value is not None:
            term, positions = self.next_value

        # Otherwise, get the next value.

        else:
            term, positions = itermerge.next(self)

        # Look at the next item to see if it has positions for the current term.

        try:
            t, p = itermerge.next(self)
            while t == term:
                positions = self.combine(positions, p)
                t, p = itermerge.next(self)
            self.next_value = t, p

        # Where an item could not be fetched, cause future requests to fail.

        except StopIteration:
            self.next_value = None

        return term, positions

    def close(self):
        for reader in self.readers:
            reader.close()
        self.readers = []

# vim: tabstop=4 expandtab shiftwidth=4
